/******************************************************************************
* Copyright (c) 2009 HarHid.labo
*
* Original code (ALivePDF Library) by Thibault Imbert.
*
* MPDF class supports to generate PDF files using Multi-Byte(Japanese).
*
* This is wrapping class for original PDF class bug. 
*
* Same as following original ALivePDF Library's license.
*
******************************************************************************/

/*
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/


package org.alivepdf.japanese.pdf
{
  import org.alivepdf.pdf.PDF;
  import org.alivepdf.pages.Page;
  import org.alivepdf.tools.sprintf;
  import flash.utils.ByteArray;

  /**
  * PDFクラスの間違いの箇所はPDFクラスを継承したPDF_mdfでバグ修正する。
  *
  */
  public class PDF_mdf extends PDF 
  {

    public function PDF_mdf ( orientation:String='Portrait', unit:String='Mm', pageSize:Object=null, rotation:int=0 ){
    }

    
  /**
  * PHPライブラリ FPDFとの相違がある箇所を修正した(１カ所).
  * 具体的にはソース参照願う。<br/>
  * <pre>
  * 修正前； if ( currentPage.w==0 ) currentPage.w = currentPage.w-this.rMargin-this.currentX;
  * 修正後： if ( width==0 ) width = currentPage.w-this.rMargin-this.currentX;
  * </pre>
  */
    override public function addCell ( width:Number=0, height:Number=0, text:String='', border:*=0, ln:Number=0, align:String='', fill:Number=0, link:String='' ):void {
      //Output a cell
      var k:Number = this.k;

      if( this.currentY + height > this.pageBreakTrigger && !this.inHeader && !this.inFooter && this.acceptPageBreak() ) {
        //Automatic page break
        var x:Number = this.currentX;
        ws=this.ws;
        if(ws>0) {
          this.ws=0;
          this.write('0 Tw');
         }
        this.addPage( new Page ( this.currentOrientation, this.defaultUnit, this.defaultSize ,currentPage.rotation ) );
        this.currentX = x;
        if(ws>0) {
          this.ws=ws;
          this.write(sprintf('%.3f Tw',ws*k));
         }
       }

/**** Here is a Bug ( next 1 line. )****/
//            if ( currentPage.w==0 ) currentPage.w = currentPage.w-this.rMargin-this.currentX;
      if ( width==0 ) width = currentPage.w-this.rMargin-this.currentX;
            
      var s:String = new String();
      var op:String;

      if( fill == 1 || border == 1 ) {
        if ( fill == 1 ) op = ( border == 1 ) ? 'B' : 'f';
        else op = 'S';
                
        s = sprintf('%.2f %.2f %.2f %.2f re %s ',this.currentX*k,(currentPage.h-this.currentY)*k,width*k,-height*k,op);
       }

      if ( border is String ) {
        currentX = this.currentX;
        currentY = this.currentY;

        var tmpBorder:String = String ( border );

        if( tmpBorder.indexOf('L') != -1 ) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',currentX*k,(currentPage.h-currentY)*k,currentX*k,(currentPage.h-(currentY+height))*k);
        if( tmpBorder.indexOf ('T') != -1) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',currentX*k,(currentPage.h-currentY)*k,(currentX+width)*k,(currentPage.h-currentY)*k);
        if( tmpBorder.indexOf ('R') != -1) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',(currentX+width)*k,(currentPage.h-currentY)*k,(currentX+width)*k,(currentPage.h-(currentY+height))*k);
        if( tmpBorder.indexOf ('B') != -1 ) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',currentX*k,(currentPage.h-(currentY+height))*k,(currentX+width)*k,(currentPage.h-(currentY+height))*k);
       }

      if ( text !== '' ) {
        var dx:Number;
        if ( align=='R' ) dx = width-this.cMargin-this.getStringWidth(text);
        else if( align=='C' ) dx = (width-this.getStringWidth(text))/2;
        else dx = this.cMargin;
        if(this.colorFlag) s+='q '+this.addTextColor+' ';
        var txt2:String = findAndReplace(')','\\)',findAndReplace('(','\\(',findAndReplace('\\','\\\\',text)));
        s+=sprintf('BT %.2f %.2f Td (%s) Tj ET',(this.currentX+dx)*k,(currentPage.h-(this.currentY+.5*height+.3*this.fontSize))*k,txt2);
        if(this.underline) s+=' '+doUnderline(this.currentX+dx,this.currentY+.5*height+.3*this.fontSize,text);
        if(this.colorFlag) s+=' Q';
        if( link ) this.addLink (this.currentX+dx,this.currentY+.5*height-.5*this.fontSize,this.getStringWidth(text),this.fontSize, link);
       }

      if ( s ) this.write(s);

      this.lasth = currentPage.h;

      if( ln >0 ) {
        //Go to next line
        this.currentY += height;
        if( ln ==1) this.currentX = this.lMargin;
       } 
      else this.currentX += width;
    }
  }
}
