/******************************************************************************
* Copyright (c) 2009 HarHid.labo
*
* MBPDF_LNX class supports to generate PDF files using Multi-Byte(Japanese),
* and this has workarounds of some bugs of Linux FlashPlayer.
*
******************************************************************************/
/*
_________________            __________________________
___    |__  /__(_)__   _________  __ \__  __ \__  ____/
__  /| |_  /__  /__ | / /  _ \_  /_/ /_  / / /_  /_    
_  ___ |  / _  / __ |/ //  __/  ____/_  /_/ /_  __/
/_/  |_/_/  /_/  _____/ \___//_/     /_____/ /_/ 
..........................................................Multi-Byte Version


Copyright (c) 2008 Thibault Imbert

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package org.alivepdf.japanese.pdf.linux {

  import org.alivepdf.pages.Page;
  import org.alivepdf.tools.sprintf;
  import flash.net.URLRequest;
  import flash.net.URLRequestHeader;
  import flash.net.URLRequestMethod;
  import flash.net.navigateToURL;
  import org.alivepdf.japanese.pdf.MBPDF;
  import org.alivepdf.japanese.metrics.MBFontMetrics;
  import org.alivepdf.japanese.metrics.MBCMap;
  import org.alivepdf.japanese.fonts.MBFontFamily;
  import org.alivepdf.images.GIFImage;
  import org.alivepdf.images.JPEGImage;
  import org.alivepdf.images.PDFImage;
  import org.alivepdf.images.PNGImage;
  import org.alivepdf.saving.Method;

  import flash.utils.ByteArray;
  import flash.system.Capabilities;

  /**
  *
  * Using MBPDF class, you can generate the PDF document including the Japanese multiByte characters.<br/>
  * 日本語Multi-Byte対応を行います。<br/>
  * <br/>
  * Linux版のFlashPlayerではいろいろ不具合が発生しますので、回避策を用意しました。<br/><br/>
  * <pre>
  * 不具合(1)
  *   PDFドキュメントを表示できない。AdobeReaderで"%PDF not found"等のエラーが出る。
  *   これはメソッドwrite内でByteArray#writeMultiByte(string,'windows_1252')をしているところが
  *   何も書き込めていない様子です。
  *   Linux版FlashPlayerの場合にはwriteMultiByteのテストを行い、NGの場合はwriteUTFBytes()を
  *   実行するように回避メソッドを用意しました(writeメソッドをオーバーライドしています)。
  *   回避策は、プロパティexecuting_Workarounds_FPLNXs_BUG = WORKAROUNDS_YES と設定された場合のみ
  *   実行されます。なお、Windows/Mac上では回避策は実行されません。
  * 不具合(2)
  *   日本語文字に変換できない。PDFドキュメント上、日本語が文字化けする。
  *   ByteArray#writeMultiByte()を使ってEncoging変換をしているのですが、Linux版のFlashPlayerでは
  *   正常に変換されません。
  *   Linux版FlashPlayerの場合にはwriteMultiByteのテストを行い、NGの場合は漢字文字コード変換テーブル
  *   を使い独自にコード変換を行います（JIS第一水準、第二水準のみ）。
  *   回避策は、プロパティexecuting_Workarounds_FPLNXs_BUG = WORKAROUNDS_YES と設定された場合のみ
  *   実行されます。なお、Windows/Mac上では回避策は実行されません。
  * 不具合(3)
  *   Web上でAlivePDFを使う場合（CGIスクリプトを用意してWEBブラウザにPDFドキュメントを表示する等）、
  *   画像をPDFドキュメントに含める、Unicodeで文字を作る等を行うと、PDFドキュメントを表示できない。
  *   Linux版のFlashPlayer-9ではPOST送信時にデータ内に非表示文字が含まれていると、データ途中で送信が
  *   途切れるようです。(FlashPlayer10ではOKっぽいです）。
  *   これを回避するために、PDFデータをBase64形式に変換したうえでPOST送信するメソッド save_Base64()
  *   を用意しました。不特定多数のLinux-Webブラウザでも利用することを想定するときは、save_Base64()を
  *   使うとよいでしょう。
  *
  * </pre>
  */
  public class MBPDF_LNX extends MBPDF {

    /**
    * @private
    */
    protected var flashPlayer_isLinux:Boolean = false;
    protected var writeMultiByte_windows1252_OK :Boolean = true;
    protected var writeMultiByte_Kanji_OK :Boolean = true;
    public static const WORKAROUNDS_YES:Boolean = true;
    /**
    * @private
    */
    protected var _executing_Workarounds_FPLNXs_BUG:Boolean = false;
    /**
    *
    * If you set WORKAROUNDS_YES to this property, MBPDF executes the workarounds of the bugs of flashplayer on Linux.
    * <br/><br/>
    * &lt;target Bug #1 &gt;<br/>
    * In Linux,FlashPlayer10, ByteArray#writeMultiByte(string,'windows_1252') writes invalid characters. <br/>   
    * So that, MBPDF(PDF) class can not generate a valid PDF file.<br/>
    * AdobeReader's error message is "%PDF not found".<br/><br/>
    * &lt;target Bug #2 &gt;<br/>
    * In Linux,FlashPlayer9 or 10, when Japanese charSet(ex.SHIFT_JIS) specified, 
    * ByteArray#writeMultiByte() writes invalid Japanese Kanji characters.
    *
    */
    public function get executing_Workarounds_FPLNXs_BUG():Boolean{
      return _executing_Workarounds_FPLNXs_BUG;
    }
    public function set executing_Workarounds_FPLNXs_BUG(value:Boolean):void {
      _executing_Workarounds_FPLNXs_BUG = value;
    }

    /**
    * The MBPDF class represents a PDF document including the Japanese multiByte characters.
    * @author HarHid  (PDF class's author:Thibault Imbert)
    * @example
    * This example shows how to create a valid PDF document :
    * <div class="listing">
    * <pre>
    * var myPDF:MBPDF = new MBPDF ( Orientation.PORTRAIT, Unit.MM, Size.A4 );
    * </pre>
    * </div>
    */
    public function MBPDF_LNX ( orientation:String='Portrait', unit:String='Mm', pageSize:Object=null, rotation:int=0 ) {
      if(MBPDF_VERSION != '0.0.1.c'){
        throw new Error ('AlivePDF Version must be 0.1.4.8 !!');
      }
      flashPlayer_isLinux = isLinux();                                     // Linux:true, others:false
      writeMultiByte_windows1252_OK = writeMultiByte_windows1252_tester(); //OK:true, NG:false
      writeMultiByte_Kanji_OK = writeMultiByte_Kanji_tester(this.charSet); //OK:true, NG:false
    }

    /**
    * Linux版のFlashPlayerであるか否かをテストします.
    * 2009年4月現在、Linux FlashPlayerでは ByteArray#writeMultiByte()に関する複数のバグが
    * 存在していますので回避させるためにLinux版FlashPlayerであるか否か知る必要があります。
    * <br/>
    * ここでは、Linux版のFlashPlayerであるか否かを確認しています。
    * @return  Linux版FlashPlayerのときTrue、以外のときFalse。
    */
    public static function isLinux():Boolean {
      var ret:Boolean = false;
      var pattern:RegExp = /(\S+)\s(\d+),(\d+),(\d+),(\d+)/i;  //ex. "LNX 10.0.26.36"
      var obj:Object = pattern.exec(Capabilities.version);
      if(obj != null){
        if(obj[1]=="LNX") {
          ret = true;
        }
      }
      return ret;
    }
    /**
    * Linux版のFlashPlayer-9であるか否かをテストします.
    * @return  Linux版FlashPlayer-9のときTrue、以外のときFalse。
    */
    public static function isLinuxFP9():Boolean {
      var ret:Boolean = false;
      var pattern:RegExp = /(\S+)\s(\d+),(\d+),(\d+),(\d+)/i;  //ex. "LNX 10.0.26.36"
      var obj:Object = pattern.exec(Capabilities.version);
      if(obj != null){
        if(obj[1]=="LNX" && obj[2] == '9') {
          ret = true;
        }
      }
      return ret;
    }
    /**
    * 以下の条件（AND条件）が成立するときに自前の文字コード変換を行うよう、オーバーライドしています。<br/>
    * Linux版FlashPlayerのとき<br/>
    * バグ回避の指定があるとき<br/>
    * ByteArray#writeMultiByte()のEncoding変換結果が不正のとき
    */
    override protected function charSetConv(s:String):String {

      /* Change the character-encoding          */
      if(flashPlayer_isLinux && executing_Workarounds_FPLNXs_BUG && !writeMultiByte_Kanji_OK) {
        return charSetConv_Linux(s);
      }
      else{
        return super.charSetConv(s);
      }
    }

    /**
    * @private
    */
    protected function charSetConv_Linux(s:String):String {

      var retString : String = '';

      for(var i:uint = 0; i<s.length; i++) {
        retString += charSetConversion(s.substr(i,1));
      }
      return retString;
    }

    /**
    * charSetの指定に応じて、UTF16のまま16進表記化して出力、またはSJIS、EUC_JP、UTF_8変換し
    * 16進表記化して出力する。
    */
    protected function charSetConversion(s1:String):String {
      var bytes:String = '';
      var c:uint = s1.charCodeAt();
      switch(characterType(s1)){
        case ASCII:
          bytes = charSetConversion_ASCII(c);
          break;
        case HALFKANA:
          bytes = charSetConversion_HalfKana(c);
          break;
        case KANJI:
          bytes = charSetConversion_Kanji(c);
          break;
        default:
          throw new Error("MBPDF_LNX#charSetConversion() something wrong.");
      }
      return bytes;
    }
    /**
    * @private
    */
    protected const ASCII:String = 'ascii';
    /**
    * @private
    */
    protected const HALFKANA:String = 'halfKana';
    /**
    * @private
    */
    protected const KANJI:String = 'kanji';
    /**
    * @private
    */
    protected function characterType(s:String):String {
      if(checkAscii(s))         
        return ASCII;
      else if(checkHalfKana(s)) 
        return HALFKANA;
      else
        return KANJI;
    }

    /**
    * @private
    */
    protected function charSetConversion_ASCII(c:uint):String{
      var bytes:String = '';
      switch(this.charSet) {
        case MBPDF.SHIFT_JIS:
        case MBPDF.EUC_JP:
          if(c==0x00a5) bytes = '5C';               // s1="\"
          else if(c==0x203e) bytes = '7E';          // s1="~"
          else bytes = uint2hex( c  & 0xff);
          break;
        case UTF_8:
          if(c==0x00a5) bytes = 'C2A5';             // s1="\"
          else if(c==0x203e) bytes = 'E280BE';      // s1="~"
          else bytes = uint2hex( c  & 0xff);
          break;
        case MBPDF.UTF_16:
          if(c==0x00a5) bytes = '005C';             // s1="\"
          else if(c==0x203e) bytes = '203E';        // s1="~"
          else bytes = '00' + uint2hex( c  & 0xff);
          break;
        default:
          throw new Error('MBPDF_LNX#harSetConversion_ASCII():charSet =('+this.charSet+')');
      }
      return bytes;
    }
    /**
    * @private
    */
    protected function charSetConversion_HalfKana(c:uint):String {
      var bytes:String = '';
      var kanaCode:uint;
      switch(this.charSet) {
        case MBPDF.SHIFT_JIS:
          kanaCode = c - 0xff60 + 0x00a0;
          bytes = uint2hex( kanaCode & 0xff );
          break;
        case MBPDF.EUC_JP:
          kanaCode = c - 0xff60 + 0x8ea0;
          bytes = uint2hex( kanaCode >>> 8 & 0xff);
          bytes += uint2hex( kanaCode & 0xff );
          break;
        case MBPDF.UTF_8:
          kanaCode = c - 0xff60 + 0xefbda0;
          bytes = uint2hex( kanaCode >>> 12 & 0xff);
          bytes += uint2hex( kanaCode >>> 8 & 0xff);
          bytes += uint2hex( kanaCode & 0xff );
          break;
        case MBPDF.UTF_16:
          kanaCode = c;
          bytes = largeValueToHex( kanaCode ); // BigEndian
          break;
        default:
          throw new Error('MBPDF_LNX#harSetConversion():charSet =('+this.charSet+')');
      }
      return bytes;
    }
    /**
    * @private
    */
    protected function charSetConversion_Kanji(c:uint):String {
      var bytes:String = '';
      var bytesBigEndian:String = largeValueToHex( c );
      if(bytesBigEndian.length==2) bytesBigEndian = "00" + bytesBigEndian;
      if(this.charSet==UTF_16)
        bytes = bytesBigEndian;
      else {
        bytes = '';
        var spaceKanji:String = '';
        var index:uint = 0;
        switch(this.charSet) {
          case SHIFT_JIS:
            spaceKanji = '8140';    // Init Kanji Space (for invalid Kanji code)
            break;
          case EUC_JP:
            spaceKanji = 'A1A1';    // Init Kanji Space (for invalid Kanji code)
            index = 1;
            break;
          case UTF_8:
            spaceKanji = 'E38080';  // Init Kanji Space (for invalid Kanji code)
            index = 2;
            break;
          default:
            throw new Error('MBPDF_LNX#harSetConversion():charSet =('+this.charSet+')');
        }
        var charsetArray:Array = CharSetTable_Kanji01.charSet_MultiBytes[bytesBigEndian];
        if(charsetArray == null){
          charsetArray = CharSetTable_Kanji02.charSet_MultiBytes[bytesBigEndian];
          if(charsetArray == null){
            bytes = spaceKanji;
          }
        }
        if(bytes == '') bytes = charsetArray[index];
      }
      return bytes;
    }

    /**
    * @private
    */
    protected function largeValueToHex(d:uint,endian:String='bigEndian'):String {
      var byte:String = '';
      var bytes:String = '';
      var w_d:uint = d;
      while(w_d) {
        byte = uint2hex(w_d & 0xff);
        if(endian=='littleEndian'){
          bytes += byte; 
        }
        else{
          bytes = byte + bytes; 
        }
        w_d = w_d >>> 8;
      }
      
      return bytes;
    }

   /**
    * PDFドキュメントデータを Base64 形式で Web-server(CGI script)へ送信する.
    * 
    * Base64の形式でPDFデータをWebサーバーに送信します。
    * Linux版FlashPlayer-9にてPOST送信をすると非表示文字の箇所(多分0x00)で送信が中断され、
    * 壊れたPDFファイルが作成されます。
    * このPDFファイルはAdobeReader等で表示できません。そこでLinux環境でもPOST送信が中断されず
    * 正しいPDFファイルが作成されるように、Base64形式で送信するようにしています。
    * <br/>
    * なお、WebサーバーのCGIスクリプトでは、Base64形式からのデコード処理をする必要があります。
    *
    * @param url The url of the CGI file
    * @param downloadMethod Lets you specify the way the PDF is going to be available. 
    * Use Download.INLINE if you want the PDF to be opened in the browser, use Download.ATTACHMENT if you want to make it available with a save-as dialog box
    * @param fileName The name of the PDF, only available when Method.REMOTE is used
    * @return null
    * @example
    * This example shows how to view the PDF in the browser with Flash or Flex :
    * <div class="listing">
    * <pre>
    *
    * if(MBPDF_LNX.isLinux()) {
    *    myPDF.save_Base64( "http://localhost/save_Base64.cgi", Download.INLINE );
    * }
    * else {
    *    myPDF.save( Method.REMOTE, "http://localhost/save.php", Download.INLINE );
    * }
    * </pre>
    * </div>
    * 
    * Base64のデコード処理をする CGI script のサンプル (Ruby)を以下に示します。
    * <div class="listing">
    * <pre>
    * require 'base64'
    * &64;in = Hash.new
    * for q in ENV['QUERY_STRING'].to_s.split(/[;&]/) do
    * key,val = q.split(/=/,2)
    *    &64;in[key] = val.gsub(/\+/," ").gsub(/%[a-fA-F\d]{2}/){ $&[1,2].hex.chr } if val
    * end
    * content_length = Integer(ENV['CONTENT_LENGTH'])
    * $stdin.binmode()
    * len = 0
    * blk = 8192
    * buffer = ''
    * while(!$stdin.eof && len < content_length)
    *    buffer += $stdin.read(blk)
    *    len += blk
    * end
    * buffer_decode = Base64.decode64(buffer)
    * print "Content-Type:application/pdf;\n\n"
    * print "Content-Length:#{buffer_decode.length}\n"
    * print 'Content-Disposition:' + "#{&64;in['method']};filename=" + '"' + "#{&64;in['name']}" + '"'+ "\n"
    * print buffer_decode
    * </pre>
    * </div>
    */
    public function save_Base64 ( url:String='', downloadMethod:String='inline', fileName:String='generated.pdf', frame:String="_blank" ):void {
      var header:URLRequestHeader = new URLRequestHeader ("Content-type","application/octet-stream");
      var myRequest:URLRequest = new URLRequest (url+'?name='+fileName+'&method='+downloadMethod );
      myRequest.requestHeaders.push (header);
      myRequest.method = URLRequestMethod.POST;
      myRequest.data = this.save(Method.BASE_64);
      navigateToURL ( myRequest, frame );
    }

    
    /**
    * Override write method to evade the bugs of Linux FlashPlayer.
    * In Linux FlashPlayer(version 10), there is a bug that ByteArray#writeMultiByte()
    * does not work correctly. To evade this bug, I have overriding the "write" method
    * to use not writeMultiByte but writeUTFBytes. <April. 2009>
    * <br/><br/>
    * 2009年4月現在、Linux FlashPlayer(version 10)では ByteArray#writeMultiByte()が
    * 正常に動作しないバグがあります。そこでwriteMultiByte()が正しく動作しないと思われる場合
    * は、writeMultiByte ではなく writeUTFBytesを使ってByteArrayに書き込むようにwriteメソッド
    * をオーバーライドしています。
    *
    */
    override protected function write( content:* ):void {
 
      if(writeMultiByte_windows1252_OK) {
        super.write(content);
      }
      else if(flashPlayer_isLinux && executing_Workarounds_FPLNXs_BUG){
        this.write_Linux( content );
      }
      else {
        super.write(content);
      }
    }

    /**
    * Linux用のwriteメソッド.
    * ByteArray#writeMultiByte()をwriteUTFBytesに変えています。
    * @parameter content PDFクラスのcontentと同じ。
    * @throw Error content PDFクラスのErrorと同じ。
    */
    protected function write_Linux( content:* ):void {
      if ( currentPage == null ) throw new Error ("No pages available, please call the addPage method first !");
      if ( state == 2 ) currentPage.content += content+"\n";
      else buffer.writeUTFBytes( content+"\n" );            //<---- changed line.
    }

    /**
    * Override createPageTree method to evade the bugs of Linux FlashPlayer.
    * In Linux FlashPlayer(version 10), there is a bug that ByteArray#writeMultiByte()
    * does not work correctly. To evade this bug, I have overriding this method
    * to use not writeMultiByte but writeUTFBytes. <April. 2009>
    * <br/><br/>
    * 2009年4月現在、Linux FlashPlayer(version 10)では ByteArray#writeMultiByte()が
    * 正常に動作しないバグがあります。そこでwriteMultiByte()が正しく動作しないと思われる場合
    * は、writeMultiByte ではなく writeUTFBytesを使ってByteArrayに書き込むようにcreatePageTree
    * メソッドをオーバーライドしています。
    *
    */
    override protected function createPageTree():void {
      if(writeMultiByte_windows1252_OK) {
        super.createPageTree();
      }
      else if(flashPlayer_isLinux && executing_Workarounds_FPLNXs_BUG){
        this.createPageTree_Linux();
      }
      else {
        super.createPageTree();
      }
    }

    /**
    * Linux用のcreatePageTreeメソッド.
    * ByteArray#writeMultiByte()をwriteUTFBytesに変えています。
    */
        protected function createPageTree_Linux():void
        {
            compressedPages = new ByteArray();

            nb = arrayPages.length;

            if( aliasNbPages != null )
            {
                for( n = 0; n<nb; n++ ) arrayPages[n].content = findAndReplace ( aliasNbPages, ( nb as String ), arrayPages[n].content );
            }

            filter = (compress) ? '/Filter /FlateDecode ' : '';

            offsets[1] = buffer.length;
            write('1 0 obj');
            write('<</Type /Pages');
            write('/Kids ['+references+']');
            write('/Count '+nb);
            write('>>');
            write('endobj');
            
            var p:String;
            var page:Page;

            for( var i:int = 0; i < nb; i++ )
            {
                page = arrayPages[i];

                newObj();
                write('<</Type /Page');
                write('/Parent 1 0 R');
                write (sprintf ('/MediaBox [0 0 %.2f %.2f]', page.width, page.height) );
                write ('/Resources 2 0 R');
                if ( page.annotations != '' ) write ('/Annots [' + page.annotations + ']');
                write ('/Rotate ' + page.rotation);
                write ('/Dur ' + 3);
                if ( page.transitions.length ) write ( page.transitions );
                write ('/Contents '+(n+1)+' 0 R>>');
                write ('endobj');
                
                if ( compress ) 
                {
//------------------compressedPages.writeMultiByte( page.content+"\n", "windows-1250" );
                    compressedPages.writeUTFBytes( page.content+"\n" );
                    compressedPages.compress();
                    compressedPages.position = 0;
                    newObj();
                    write('<<'+filter+'/Length '+compressedPages.length+'>>');
                    write('stream');
                    buffer.writeBytes( compressedPages );
                    buffer.writeUTFBytes("\n");
                    write('endstream');
                    write('endobj');                    
                } else 
                {
                    newObj();
                    write('<<'+filter+'/Length '+page.content.length+'>>');
                    writeStream(page.content.substr(0, page.content.length-1));
                    write('endobj');
                }
            }
        }




    /**
    * writeMultiByte_ASCII_checkerを呼び出して、ByteArray#writeMultiByte(string,'windows_1252')が
    * 正しく動作するかをテストします。
    */
    protected function writeMultiByte_windows1252_tester():Boolean {
      return writeMultiByte_ASCII_checker('windows_1252');
    }
    /**
    * ByteArray#writeMultiByte()により漢字のエンコード変換が正しく行われるかテストします.
    * 2009年4月現在、Linux FlashPlayerでは ByteArray#writeMultiByte()が
    * 正常に動作しないバグがあるため、writeMultiByteが正常に動作するか否かを確認するために用意しています。
    * <br/>
    * 代表の漢字２文字「申」「乳」を指定されたcharSetに応じてエンコード変換(writeMultiByteを使って)
    * し、結果が仕様どおりか否かを確認しています。
    *
    */
    protected function writeMultiByte_Kanji_tester(cs:String):Boolean {
      var ret:Boolean = false;
      var hex1:String = '';
      var hex2:String = '';
      switch(cs) {
        case MBPDF.SHIFT_JIS:
          hex1 = '905C'; // '申'
          hex2 = '93FB'; // '乳'
          break;
        case MBPDF.EUC_JP:
          hex1 = 'BFBD'; // '申'
          hex2 = 'C6FD'; // '乳'
          break;
        case MBPDF.UTF_8:
          hex1 = 'E794B3'; // '申'
          hex2 = 'E4B9B3'; // '乳'
//          throw new Error('MBPDF#writeMultiByte_Kanji_tester():UTF8への変換は現在対応していません。');
          break;
        case MBPDF.UTF_16:
          return true;
        default:
          throw new Error('MBPDF#writeMultiByte_Kanji_tester():Error charSet ='+cs);
      }
      if(writeMultiByte_Kanji_checker(cs,'申',hex1) &&
         writeMultiByte_Kanji_checker(cs,'乳',hex2)) {
         ret = true;
      }
      return ret;
    }
    /**
    * @private
    */
    protected function writeMultiByte_Kanji_checker(cs:String,kanji:String,hex:String):Boolean {
      var ret:Boolean = false;
      var bf:ByteArray = new ByteArray();
      bf.writeMultiByte(kanji,cs);
      if(readByte_HexString(bf) == hex) ret = true;
      return ret;
    }

    /**
    * ByteArray#writeMultiByte()によるASCII文字の書き込みが正しく行われるかテストします.
    * 2009年4月現在、Linux FlashPlayer１０では ByteArray#writeMultiByte()が
    * 正常に動作しないバグがあるため、writeMultiByteが正常に動作するか否かを確認するために用意しました。
    * <br/>
    * 代表のASCII文字 '1','a'の２文字を指定されたcharSetに応じてエンコード変換(writeMultiByteを使って)
    * し、結果が仕様どおりか否かを確認しています。
    *
    */
    protected function writeMultiByte_ASCII_checker(cs:String):Boolean {

      var ret:Boolean = false;
      var bf_1:ByteArray = new ByteArray();
      var bf_a:ByteArray = new ByteArray();
      bf_1.writeMultiByte('1',cs);
      bf_a.writeMultiByte('a',cs);

      trace('writeMultiByte_ASCII_checker(' + cs + ')');
      trace('writeMultiByte_ASCII_checker readByte_HexString(bf_1)=' + readByte_HexString(bf_1));
      trace('writeMultiByte_ASCII_checker readByte_HexString(bf_a)=' + readByte_HexString(bf_a));

      if(readByte_HexString(bf_1)=='31' && readByte_HexString(bf_a)=='61'){ 
        //２文字が合致すればOKとみなす。
        ret = true;
      }
      return ret;
    }
    
    /**
    * charSetのsetter内で、writeMultiByte_Kanji_tester()のチェックを行うようにオーバーライドする。
    *
    */
    override public function set charSet(value:String):void {
      super.charSet = value;
      writeMultiByte_Kanji_OK = (value == MBPDF.UTF_16)? false:writeMultiByte_Kanji_tester(value);
    }
  }
}

